#include "PublicKey.hpp"

void PublicKey::encrypt(CurveBipoint& G_element, const Scalar& cleartext) const
{
    Scalar lambda;
    lambda.set_random();

    CurveBipoint cleartext_as_element, random_mask;
    cleartext_as_element = get_bipoint_curvegen() * cleartext;
    random_mask = get_bipoint_curve_subgroup_gen() * lambda;

    G_element = cleartext_as_element + random_mask;
}

void PublicKey::encrypt(TwistBipoint& H_element, const Scalar& cleartext) const
{
    Scalar lambda;
    lambda.set_random();

    TwistBipoint cleartext_as_element, random_mask;
    cleartext_as_element = get_bipoint_twistgen() * cleartext;
    random_mask = get_bipoint_twist_subgroup_gen() * lambda;

    H_element = cleartext_as_element + random_mask;
}

void PublicKey::encrypt(CurveBipoint& G_element, TwistBipoint& H_element, const Scalar& cleartext) const
{
    encrypt(G_element, cleartext);
    encrypt(H_element, cleartext);
}

CurveBipoint PublicKey::homomorphic_addition(const CurveBipoint& a, const CurveBipoint& b) const
{
    Scalar lambda;
    lambda.set_random();

    CurveBipoint random_mask;
    random_mask = bipoint_curve_subgroup_gen * lambda;

    return a + b + random_mask;
}

TwistBipoint PublicKey::homomorphic_addition(const TwistBipoint& a, const TwistBipoint& b) const
{
    Scalar lambda;
    lambda.set_random();

    TwistBipoint random_mask;
    random_mask = bipoint_twist_subgroup_gen * lambda;

    return a + b + random_mask;
}

Quadripoint PublicKey::homomorphic_addition(const Quadripoint& a, const Quadripoint& b) const
{
    Quadripoint random_mask;
    CurveBipoint random_mask_curve;
    TwistBipoint random_mask_twist;
    
    Scalar lambda1, lambda2;
    lambda1.set_random();
    lambda2.set_random();

    random_mask_curve = bipoint_curve_subgroup_gen * lambda1;
    random_mask_twist = bipoint_twist_subgroup_gen * lambda2;
    random_mask = pairing(bipoint_curvegen, random_mask_twist) + pairing(random_mask_curve, bipoint_twistgen);

    return a + b + random_mask;
}

Quadripoint PublicKey::homomorphic_multiplication(const CurveBipoint& a, const TwistBipoint& b) const
{
    Quadripoint random_mask;
    CurveBipoint random_mask_curve;
    TwistBipoint random_mask_twist;
    
    Scalar lambda1, lambda2;
    lambda1.set_random();
    lambda2.set_random();

    random_mask_curve = bipoint_curve_subgroup_gen * lambda1;
    random_mask_twist = bipoint_twist_subgroup_gen * lambda2;
    random_mask = pairing(bipoint_curvegen, random_mask_twist) + pairing(random_mask_curve, bipoint_twistgen);

    return pairing(a, b) + random_mask;
}

CurveBipoint PublicKey::get_bipoint_curvegen() const
{
    return bipoint_curvegen;
}

TwistBipoint PublicKey::get_bipoint_twistgen() const
{
    return bipoint_twistgen;
}

CurveBipoint PublicKey::get_bipoint_curve_subgroup_gen() const
{
    return bipoint_curve_subgroup_gen;
}

TwistBipoint PublicKey::get_bipoint_twist_subgroup_gen() const
{
    return bipoint_twist_subgroup_gen;
}

PublicKey::PublicKey()
{ }

void PublicKey::set(const CurveBipoint& g, const TwistBipoint& h, const CurveBipoint& g1, const TwistBipoint& h1)
{
    bipoint_curvegen = g;
    bipoint_twistgen = h;
    
    bipoint_curve_subgroup_gen = g1;
    bipoint_twist_subgroup_gen = h1;
}
